require 'socket'
require 'logger'
require 'securerandom'
require 'active_model'

require 'rubypitaya/core/path'
require 'rubypitaya/core/setup'
require 'rubypitaya/core/config'
require 'rubypitaya/core/session'
require 'rubypitaya/core/postman'
require 'rubypitaya/core/parameters'
require 'rubypitaya/core/http_routes'
require 'rubypitaya/core/routes_base'
require 'rubypitaya/core/status_codes'
require 'rubypitaya/core/handler_router'
require 'rubypitaya/core/etcd_connector'
require 'rubypitaya/core/nats_connector'
require 'rubypitaya/core/database_config'
require 'rubypitaya/core/instance_holder'
require 'rubypitaya/core/mongo_connector'
require 'rubypitaya/core/redis_connector'
require 'rubypitaya/core/database_connector'
require 'rubypitaya/core/initializer_content'
require 'rubypitaya/core/initializer_broadcast'
require 'rubypitaya/core/application_files_importer'

module RubyPitaya

  class Main

    attr_reader :redis_connector, :config, :bll

    def initialize
      @setup = Setup.new
      @environment_name = @setup.fetch('rubypitaya.server.environment', 'development')
      @is_development_environment = @environment_name == 'development'
      @setup.auto_reload if @is_development_environment

      @log = Logger.new('/proc/self/fd/1')
      @log.level = Logger::INFO
      @log.formatter = proc do |severity, datetime, progname, msg|
        "#{msg}\n"
      end

      @application_files_importer = ApplicationFilesImporter.new
      @application_files_importer.import
      @application_files_importer.auto_reload if @is_development_environment

      @server_name = @setup['rubypitaya.server.name']
      @service_uuid = SecureRandom.uuid
      @desktop_name = Socket.gethostname

      @etcd_prefix = @setup['rubypitaya.etcd.prefix']
      @etcd_address = @setup['rubypitaya.etcd.url']
      @etcd_lease_seconds = @setup['rubypitaya.etcd.leaseSeconds'].to_i
      @allow_reconnect = false
      @etcd_connector = EtcdConnector.new(@service_uuid, @desktop_name,
                                          @server_name, @etcd_prefix,
                                          @etcd_address, @allow_reconnect,
                                          @etcd_lease_seconds)
      @etcd_connector.connect

      @nats_address = @setup['rubypitaya.nats.url']
      @nats_connector = NatsConnector.new(@nats_address, @service_uuid,
                                          @server_name)

      @redis_address = @setup['rubypitaya.redis.url']
      @redis_connector = RedisConnector.new(@redis_address)
      @redis_connector.connect

      @database_config = DatabaseConfig.new
      @database_connector = DatabaseConnector.new(@database_config)
      @database_connector.connect

      @mongo_address = @setup['rubypitaya.mongo.url']
      @mongo_user = @setup['rubypitaya.mongo.user']
      @mongo_password = @setup['rubypitaya.mongo.pass']
      @mongo_database_name = @setup['rubypitaya.mongo.database']
      @mongo_connector = MongoConnector.new(@mongo_address, @mongo_user,
                                            @mongo_password, @mongo_database_name)
      @mongo_connector.connect

      @session = Session.new
      @postman = Postman.new(@nats_connector)
      @config = Config.new
      @config.auto_reload if @is_development_environment

      @bll = InstanceHolder.new

      @initializer_content = InitializerContent.new(@log,
                                                    @redis_connector.redis,
                                                    @mongo_connector.mongo,
                                                    @setup,
                                                    @config)
      @initializer_broadcast = InitializerBroadcast.new
      @initializer_broadcast.run(@initializer_content)

      @handler_router = HandlerRouter.new()

      run_http
      run_server
    end

    private

    def run_http
      HttpRoutes.set :bll, @bll
      HttpRoutes.set :log, @log
      HttpRoutes.set :redis, @redis_connector.redis
      HttpRoutes.set :mongo, @mongo_connector.mongo
      HttpRoutes.set :setup, @setup
      HttpRoutes.set :config, @config
      HttpRoutes.set :views, [Path::HTTP_VIEWS_PATH] + Path::Plugins::HTTP_VIEWS_PATHS

      HttpRoutes.auto_reload if @is_development_environment

      Thread.new do
        HttpRoutes.bind = '0.0.0.0'
        HttpRoutes.port = '4567'
        HttpRoutes.run!
      end
    end

    def run_server
      @is_shutting_down = false

      catch :sig_shutdown do
        Signal.trap("SIGINT") { throw :sig_shutdown }
        Signal.trap("SIGQUIT") { throw :sig_shutdown }
        Signal.trap("SIGTERM") { throw :sig_shutdown }

        @log.info "Server started!"
        run_nats_connection
      end

      Signal.trap("SIGINT", nil)
      Signal.trap("SIGQUIT", nil)
      Signal.trap("SIGTERM", nil)

      shutdown_server
    end

    def shutdown_server
      return if @is_shutting_down
      @is_shutting_down = true

      @log.info "Server shutting down..."

      @etcd_connector.disconnect
      @database_connector.disconnect

      exit(0)
    end

    def run_nats_connection
      @nats_connector.connect do |request|
        start_time_seconds = Time.now.to_f

        message_route = request[:msg][:route]
        message_data = request[:msg][:data] || {}
        message_data = JSON.parse(message_data) if message_data.class == String

        params = Parameters.new(message_data)

        session_id = request[:session][:id]
        session_uid = request[:session].fetch(:uid, '')
        session_data = request[:session][:data]
        session_data = JSON.parse(session_data, symbolize_names: true)
        frontend_id = request[:frontendID]
        metadata = request[:metadata]
        metadata = JSON.parse(metadata, symbolize_names: true)

        @session.update(session_id, session_uid, session_data, metadata,
                        frontend_id)

        # TODO: Validate if handler and action names are not nil
        handler_name, action_name = message_route.scan(/\A\w+\.(\w+)\.(\w+)\z/)[0]

        @log.info "request  -> route: #{message_route}"
        @log.info "         -> data:  #{message_data}"

        @config.clear_cache

        response = @handler_router.call(handler_name, action_name, @session,
                                        @postman, @redis_connector.redis,
                                        @mongo_connector.mongo, @setup, @config,
                                        @log, params)

        delta_time_seconds = ((Time.now.to_f - start_time_seconds) * 1000).round(2)

        @log.info "response [#{delta_time_seconds}ms] -> #{response.to_json}"

        response
      end
      rescue Exception => error
        @log.info "ERROR: #{error}"
        @log.info error.backtrace.join("\n")
        run_nats_connection
    end
  end
end
