require 'protobuf'
require 'nats/client'

module RubyPitaya

  class NatsConnector

    REQUEST_TYPE_SYS  = 0
    REQUEST_TYPE_USER = 1

    MESSAGE_TYPE_REQUEST  = 0
    MESSAGE_TYPE_NOTIFY   = 1
    MESSAGE_TYPE_RESPONSE = 2
    MESSAGE_TYPE_PUSH     = 3

    def initialize(nats_address, server_uuid, server_name)
      @server_uuid = server_uuid
      @server_name = server_name
      @nats_address = nats_address
    end

    def connect
      NATS.start(:servers => [@nats_address]) do |nc|

        NATS.subscribe(subscribe_topic) do |message, reply, a, b|
          request = NatsRequest.decode(message).to_hash

          Fiber.new do
            response = yield request

            nats_response = NatsResponse.new(data: response.to_json)

            NATS.publish(reply, nats_response)
          end.resume
        end

      end
    end

    def push_to_frontend(session, message_route, payload)
      frontend_topic = get_frontend_topic(session.frontend_id)

      nats_message = NatsMessage.new(
        route: message_route,
        data: payload,
        reply: @server_uuid,
        type: MESSAGE_TYPE_REQUEST,
      )

      nats_session = NatsSession.new(
        id: session.id,
        uid: session.uid,
        data: session.data.to_json,
      )

      nats_request = NatsRequest.new(
        type: REQUEST_TYPE_USER,
        session: nats_session,
        msg: nats_message,
        frontendID: session.frontend_id,
        metadata: session.metadata.to_json,
      )

      nats_response = NATS.request(frontend_topic, nats_request)
      response = NatsResponse.decode(nats_response).to_hash

      response
    end

    def push_to_user(uid, message_route, payload)
      user_topic = get_user_message_topic(uid)

      nats_push = NatsPush.new(
        route: message_route,
        uid: uid,
        data: payload.to_json,
      )

      NATS.publish(user_topic, nats_push)

      0
    end

    def subscribe_topic
      "pitaya/servers/#{@server_name}/#{@server_uuid}"
    end

    def get_frontend_topic(frontend_id)
      "pitaya/servers/connector/#{frontend_id}"
    end

    def get_user_message_topic(uid)
      "pitaya/connector/user/#{uid}/push"
    end
  end

  class NatsSession < ::Protobuf::Message; end
  class NatsMessage < ::Protobuf::Message; end
  class NatsRequest < ::Protobuf::Message; end
  class NatsResponse < ::Protobuf::Message; end
  class NatsError < ::Protobuf::Message; end
  class NatsBindMsg < ::Protobuf::Message; end
  class NatsPush < ::Protobuf::Message; end

  class NatsSession
    optional :int32, :id, 1
    optional :bytes, :uid, 2
    optional :bytes, :data, 3
  end

  class NatsMessage
    optional :int32, :id, 1
    optional :bytes, :route, 2
    optional :bytes, :data, 3
    optional :bytes, :reply, 4
    optional :int32, :type,  5
  end

  class NatsRequest
    optional :int32, :type, 1
    optional NatsSession, :session, 2
    optional NatsMessage, :msg, 3
    optional :bytes, :frontendID, 4
    optional :bytes, :metadata, 5
  end

  class NatsError
    optional :bytes, :code, 1
    optional :bytes, :message, 2
    optional :bytes, :metadata, 3
  end

  class NatsResponse
    optional :bytes, :data, 1
    optional NatsError, :error, 2
  end

  class NatsBindMsg
    optional :bytes, :uid, 1
    optional :bytes, :fid, 2
  end

  class NatsPush
    optional :bytes, :route, 1
    optional :bytes, :uid, 2
    optional :bytes, :data, 3
  end
end
